
/*   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2 only.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   (C) Kent E. Yoder, 2006
 */

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include <openssl/rand.h>
#include <openssl/rsa.h>
#include <openssl/aes.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/err.h>

#include <trousers/tss.h>
#include <trousers/trousers.h>

#include "tpm_keyring_util.h"

#define TPM_E_NON_FATAL12	0x800
#define TPM_E_NON_FATAL11	0x400


const TSS_UUID SRK_UUID = TSS_UUID_SRK;

void
OpenSSL_init()
{
	OpenSSL_add_all_algorithms();
	ERR_load_ERR_strings();
	ERR_load_crypto_strings();
}

void
openssl_print_errors()
{
	ERR_print_errors_fp(stderr);
}

char *
generate_swkey(tpm_keyring_util_ctx *ctx)
{
	if (ctx->bits == 0)
		ReturnError("Cannot generate a key of size 0!");

	if ((ctx->sw_key = RSA_generate_key(ctx->bits, 65537, NULL, NULL)) == NULL)
		ReturnError("Failed generating software key!");

	return NULL;
}

char *
util_bytes_to_string(char *bytes, int chars)
{
	char *ret = (char *)malloc((chars*2) + 1);
	int i, len = chars*2;

	if (ret == NULL)
		return ret;

	for (i = 0; i < chars; i+=4) {
		sprintf(&ret[i*2], "%02x%02x%02x%02x", bytes[i] & 0xff,
				bytes[i+1] & 0xff, bytes[i+2] & 0xff,
				bytes[i+3] & 0xff);
	}

	ret[len] = '\0';

	return ret;
}

TSS_RESULT
generate_uuid(char *group_name, TSS_UUID *uuid)
{
	BYTE hash[SHA1_HASH_SIZE];
	EVP_MD_CTX md_ctx;

	EVP_MD_CTX_init(&md_ctx);
	if (!EVP_DigestInit_ex(&md_ctx, EVP_sha1(), NULL))
		return 1;
	if (!EVP_DigestUpdate(&md_ctx, group_name, strlen(group_name)))
		return 1;
	if (!EVP_DigestFinal(&md_ctx, hash, NULL))
		return 1;

	memcpy(uuid, hash, sizeof(TSS_UUID));

	return 0;
}

char *
tpm_keyring_util_get_uuid(char *name)
{
	TSS_UUID uuid;

	generate_uuid(name, &uuid);

	return util_bytes_to_string((char *)&uuid, sizeof(TSS_UUID));
}

int
util_set_file_mode(char *filename, mode_t mode)
{
	struct stat file_stat;

	if (stat(filename, &file_stat) == -1) {
		LogDebug("%s: stat: %s", __FUNCTION__, strerror(errno));
		return -1;
	} else if ((file_stat.st_mode ^ mode) != 0) {
		if (chmod(filename, mode) == -1) {
			LogDebug("chmod(%s) failed: %s", filename, strerror(errno));
			return -1;
		}
	}

	return 0;
}

int
openssl_write_key(RSA *rsa, char *filename, char *p)
{
	BIO *b = BIO_new_file(filename, "w");
	int p_len;
	EVP_CIPHER *cipher = (EVP_CIPHER *)EVP_aes_256_cbc();

	if (!b) {
		LogDebug("%s: Error opening file for write: %s", __FUNCTION__, filename);
		return -1;
	}

	if (p && strlen(p) > 0) {
		LogDebug("Password used to save: %s, strlen(p)=%d", p, strlen(p));
		p_len = strlen(p);
	} else {
		cipher = NULL;
		p_len = 0;
	}

	if (!PEM_write_bio_RSAPrivateKey(b, rsa, (const EVP_CIPHER *)cipher, (BYTE *)p, p_len, 0, NULL)) {
		BIO_free(b);
		LogDebug("Writing key %s to disk failed.", filename);
		DEBUG_openssl_print_errors();
		return -1;
	}

	BIO_free(b);

	if (util_set_file_mode(filename, (S_IRUSR|S_IWUSR))) {
		LogDebug("Setting file mode of %s failed", filename);
	}

	return 0;
}

char *
save_swkey(tpm_keyring_util_ctx *ctx)
{
	char *ascii_hash, buf[4096];
	TSS_UUID uuid;

	if (generate_uuid(ctx->name, &uuid))
		ReturnError("Out of memory");

	ascii_hash = util_bytes_to_string((char *)&uuid, sizeof(TSS_UUID));
	if (ascii_hash == NULL)
		ReturnError("Out of memory");

	snprintf(buf, 4096, "%s/%s.pem", ctx->path, ascii_hash);
	LogDebug("Group %s hashes to: %s", ctx->name, ascii_hash);
	free(ascii_hash);

	if (openssl_write_key(ctx->sw_key, buf, ctx->pass))
		ReturnError("openssl_write_key failed");

	return NULL;
}

char *
tpm_keyring_util_remove_swkey(char *name, char *path)
{
	char *ascii_hash, buf[4096];
	TSS_UUID uuid;
	int rc;

	OpenSSL_init();

	if (generate_uuid(name, &uuid))
		ReturnError("Out of memory");

	ascii_hash = util_bytes_to_string((char *)&uuid, sizeof(TSS_UUID));
	if (ascii_hash == NULL)
		ReturnError("Out of memory");

	snprintf(buf, 4096, "%s/%s.pem", path, ascii_hash);
	LogDebug("Group %s hashes to: %s", name, ascii_hash);
	free(ascii_hash);

	if ((rc = unlink(buf)) != 0)
		ReturnError("Deletion of %s failed: %s", strdup(buf), strerror(errno));

	return NULL;
}

RSA *
openssl_read_key(char *filename, char *p)
{
	BIO *b;
	RSA *rsa = NULL;

	b = BIO_new_file(filename, "r");
	if (b == NULL) {
		LogDebug("%s: Error opening file for read: %s", __FUNCTION__, filename);
		return NULL;
	}

	if (p && strlen(p) > 0) {
		LogDebug("Password used to save: %s", p);
	} else {
		return (RSA *)0x1;
	}

	if ((rsa = PEM_read_bio_RSAPrivateKey(b, NULL, 0, (void *)p)) == NULL) {
		BIO_free(b);
		LogDebug("Reading key %s from disk failed.", filename);
		DEBUG_openssl_print_errors();
		return (RSA *)0x1;
	}

	BIO_free(b);

	return rsa;
}

char *
load_swkey(tpm_keyring_util_ctx *ctx)
{
	char *ascii_hash, buf[4096];
	TSS_UUID uuid;

	if (generate_uuid(ctx->name, &uuid))
		ReturnError("Out of memory");

	ascii_hash = util_bytes_to_string((char *)&uuid, sizeof(TSS_UUID));
	if (ascii_hash == NULL)
		ReturnError("Out of memory");

	snprintf(buf, 4096, "%s/%s.pem", ctx->path, ascii_hash);
	LogDebug("Group %s hashes to: %s", ctx->name, ascii_hash);
	free(ascii_hash);

	if ((ctx->sw_key = openssl_read_key(buf, ctx->pass)) == NULL)
		ReturnError("Cannot find software key %s.\n"
			    "Perhaps you moved it into safe keeping?", buf);
	else if (ctx->sw_key == (RSA *)0x1)
		ReturnError("swprompt");

	return NULL;
}

int key_sizes[] = { 512, 1024, 2048, 0 };

int
key_size_is_valid(int key_size)
{
	int i = 0;

	while (key_sizes[i] != 0) {
		if (key_size == key_sizes[i])
			return 0;
		i++;
	}

	return -1;
}

void
tpm_keyring_util_ctx_close(tpm_keyring_util_ctx *ctx)
{
	Tspi_Context_Close(ctx->hContext);
	if (ctx->sw_key)
		RSA_free(ctx->sw_key);
}

tpm_keyring_util_ctx *
tpm_keyring_util_create_ctx(int bits, char *name, char *pass, char *host, char *path)
{
	tpm_keyring_util_ctx *myself = calloc(1, sizeof(tpm_keyring_util_ctx));
	if (myself == NULL) {
		LogDebug("malloc of %zd bytes failed.", sizeof(tpm_keyring_util_ctx));
		return NULL;
	}

	myself->bits = bits;
	myself->name = name;

	if (pass && strlen(pass) == 0)
		myself->pass = NULL;
	else
		myself->pass = pass;

	if (host && strlen(host) == 0)
		myself->host = NULL;
	else
		myself->host = host;

	myself->path = path;

	return myself;
}

char *
init_tss(tpm_keyring_util_ctx *ctx)
{
	TSS_RESULT result;
	UNICODE *uHost = NULL;
	TSS_HPOLICY hPolicy;
	UINT32 size;

	result = Tspi_Context_Create(&ctx->hContext);
	if (result)
                ReturnError("Tspi_Context_Create failed: %s", Trspi_Error_String(result));

	if (ctx->host) {
		size = strlen(ctx->host) + 1;
		uHost = (UNICODE *)Trspi_Native_To_UNICODE((BYTE *)ctx->host, &size);
		if (uHost == NULL) {
			Tspi_Context_Close(ctx->hContext);
			ReturnError("Error converting host name \"%s\" to unicode.", ctx->host);
		}
	}

	result = Tspi_Context_Connect(ctx->hContext, uHost);
        if (result) {
		Tspi_Context_Close(ctx->hContext);
                ReturnError("Tspi_Context_Connect failed: %s", Trspi_Error_String(result));
	}

	result = Tspi_Context_LoadKeyByUUID(ctx->hContext, TSS_PS_TYPE_SYSTEM, SRK_UUID,
					    &ctx->hSRK);
	if (result)
		ReturnError("Tspi_Context_LoadKeyByUUID failed: %s", Trspi_Error_String(result));

	result = Tspi_GetPolicyObject(ctx->hSRK, TSS_POLICY_USAGE, &hPolicy);
	if (result)
		ReturnError("Tspi_GetPolicyObject failed: %s", Trspi_Error_String(result));

	result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN, 0, NULL);
	if (result)
		ReturnError("Tspi_Policy_SetSecret failed: %s", Trspi_Error_String(result));

	return NULL;
}

UINT32
get_tss_key_size(unsigned bits)
{
	UINT32 rv;

	switch (bits) {
		case 512:
			rv = TSS_KEY_SIZE_512;
			break;
		case 1024:
			rv = TSS_KEY_SIZE_1024;
			break;
		case 2048:
			rv = TSS_KEY_SIZE_2048;
			break;
		default:
			rv = 0;
			break;
	}

	return rv;
}


int
openssl_get_modulus_and_prime(RSA *rsa, unsigned int *size_n, unsigned char *n,
		unsigned int *size_p, unsigned char *p)
{
	/* get the modulus from the RSA object */
	if ((*size_n = BN_bn2bin(rsa->n, n)) <= 0) {
		DEBUG_openssl_print_errors();
		return -1;
	}

	/* get one of the primes from the RSA object */
	if ((*size_p = BN_bn2bin(rsa->p, p)) <= 0) {
		DEBUG_openssl_print_errors();
		return -1;
	}

	return 0;
}

char *
wrap(tpm_keyring_util_ctx *ctx, char *pass)
{
	unsigned char n[256], p[256];
	int size_n, size_p;
	TSS_RESULT result;
	TSS_HPOLICY hPolicy;
	UINT32 pubKeyLen, tss_key_size, uPassLen, pass_len = 0;
	BYTE *pubKey, *uPass = NULL, *pass_ptr = NULL;
	UINT32 num_retries = 1;

	if (ctx->sw_key == NULL)
		ReturnError("Error: There's no software key to wrap!");

	if (pass) {
		/* uPass will be used if auth fails for the ascii password */
		uPassLen = strlen(pass);
		uPass = Trspi_Native_To_UNICODE((BYTE *)pass, &uPassLen);

		pass_ptr = (BYTE *)pass;
		pass_len = strlen(pass);

		result = Tspi_GetPolicyObject(ctx->hSRK, TSS_POLICY_USAGE, &hPolicy);
		if (result)
			ReturnError("Tspi_GetPolicyObject failed: %s", Trspi_Error_String(result));
	}

try_new_auth:
	if (pass_ptr) {
		result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN, pass_len, pass_ptr);
		if (result)
			ReturnError("Tspi_Policy_SetSecret failed: %s", Trspi_Error_String(result));
	}
retry:
	if ((result = Tspi_Key_GetPubKey(ctx->hSRK, &pubKeyLen, &pubKey))) {
		if (num_retries < 5 &&
		    (result & TPM_E_NON_FATAL11 || result & TPM_E_NON_FATAL12)) {
			/* The TPM is denying our request since too many auth's are failing. Just
			 * retry after a timeout of 1/4 a second */
			usleep(250000 * num_retries++);
			goto retry;
		}

		if (Trspi_Error_Code(result) == TCPA_E_AUTHFAIL) {
			/* if we just tried the ascii password, retry using the unicode password,
			 * else we're in a failure state where we need a new password to try */
			if (pass && (pass_ptr == (BYTE *)pass)) {
				pass_ptr = uPass;
				pass_len = uPassLen;
				goto try_new_auth;
			}
			ReturnError("prompt");
		}

		if (Trspi_Error_Code(result) == TCPA_E_KEYNOTFOUND)
			ReturnError("nosrk");

		ReturnError("Tspi_Key_GetPubKey failed: 0x%x, %s", result,
			    Trspi_Error_String(result));
	}

	Tspi_Context_FreeMemory(ctx->hContext, pubKey);

	if ((tss_key_size = get_tss_key_size(RSA_size(ctx->sw_key) * 8)) == 0)
		ReturnError("Bad software key size: %d", RSA_size(ctx->sw_key) * 8);

	/* create the TSS key object */
	result = Tspi_Context_CreateObject(ctx->hContext, TSS_OBJECT_TYPE_RSAKEY,
					   tss_key_size | TSS_KEY_TYPE_BIND | TSS_KEY_MIGRATABLE |
					   TSS_KEY_NO_AUTHORIZATION, &ctx->hKey);
	if (result)
		ReturnError("Tspi_Context_CreateObject failed: %s", Trspi_Error_String(result));

	if (openssl_get_modulus_and_prime(ctx->sw_key,
					  (UINT32 *)&size_n, n, (UINT32 *)&size_p, p) != 0)
		ReturnError("Error extracting modulus and primes!");

	/* set the public key data in the TSS object */
	result = Tspi_SetAttribData(ctx->hKey, TSS_TSPATTRIB_RSAKEY_INFO,
				    TSS_TSPATTRIB_KEYINFO_RSA_MODULUS, size_n, n);
	if (result)
		ReturnError("Tspi_SetAttribData failed: %s", Trspi_Error_String(result));

	/* set the private key data in the TSS object */
	result = Tspi_SetAttribData(ctx->hKey, TSS_TSPATTRIB_KEY_BLOB,
				    TSS_TSPATTRIB_KEYBLOB_PRIVATE_KEY, size_p, p);
	if (result)
		ReturnError("Tspi_SetAttribData failed: %s", Trspi_Error_String(result));

	result = Tspi_Key_WrapKey(ctx->hKey, ctx->hSRK, 0);
	if (result)
		ReturnError("Tspi_Key_WrapKey failed: %s", Trspi_Error_String(result));

	return NULL;
}

char *
tpm_keyring_util_send_key_retry(tpm_keyring_util_ctx *ctx, char *srkpass)
{
	char *ret;
	TSS_RESULT result;
	TSS_UUID uuid;

	if ((result = generate_uuid(ctx->name, &uuid)))
		ReturnError("Failed to generate a new uuid: %s", Trspi_Error_String(result));

	if ((ret = wrap(ctx, srkpass)))
		return ret;

	if ((result = Tspi_Context_RegisterKey(ctx->hContext, ctx->hKey, TSS_PS_TYPE_SYSTEM, uuid,
					       TSS_PS_TYPE_SYSTEM, SRK_UUID)))
		ReturnError("Failed to register key in system persistent storage: %s",
			    Trspi_Error_String(result));

	Tspi_Context_Close(ctx->hContext);

	return NULL;
}

char *
tpm_keyring_util_send_key(tpm_keyring_util_ctx *ctx)
{
	TSS_RESULT result;
	TSS_UUID uuid;
	char *ret;

	OpenSSL_init();

	if ((result = generate_uuid(ctx->name, &uuid)))
		ReturnError("Failed to generate a new uuid: %s", Trspi_Error_String(result));

	if ((ret = load_swkey(ctx)))
		return ret;

	if ((ret = init_tss(ctx)))
		return ret;

	if ((ret = wrap(ctx, NULL)))
		return ret;

	if ((result = Tspi_Context_RegisterKey(ctx->hContext, ctx->hKey, TSS_PS_TYPE_SYSTEM, uuid,
					       TSS_PS_TYPE_SYSTEM, SRK_UUID)))
		ReturnError("Failed to register key: %s", Trspi_Error_String(result));

	Tspi_Context_Close(ctx->hContext);

	return NULL;
}

char *
tpm_keyring_util_unregister_local(char *name)
{
	TSS_HCONTEXT hContext;
	TSS_HKEY hKey;
	TSS_RESULT result;
	TSS_UUID uuid;

	if ((result = generate_uuid(name, &uuid)))
		ReturnError("Error: Failed to generate uuid!");

	if ((result = Tspi_Context_Create(&hContext)))
                ReturnError("Tspi_Context_Create failed: %s", Trspi_Error_String(result));

	LogDebug("Unregistered key from user PS");
	if ((result = Tspi_Context_UnregisterKey(hContext, TSS_PS_TYPE_USER, uuid, &hKey))) {
		Tspi_Context_Close(hContext);
		ReturnError("Error unregistering key from persistent storage: %s",
			    Trspi_Error_String(result));
	}

	Tspi_Context_Close(hContext);

	return NULL;
}

int
tpm_keyring_util_get_keysize(char *name)
{
	TSS_HCONTEXT hContext;
	TSS_HKEY hKey;
	TSS_RESULT result;
	TSS_UUID uuid;
	UINT32 keysize;

	/* generate the UUID based on the group name */
	if ((result = generate_uuid(name, &uuid)))
		return 0;

	if ((result = Tspi_Context_Create(&hContext)))
		return 0;

	if ((result = Tspi_Context_GetKeyByUUID(hContext, TSS_PS_TYPE_USER, uuid, &hKey)))
		return 0;

	if ((result = Tspi_GetAttribUint32(hKey, TSS_TSPATTRIB_KEY_INFO, TSS_TSPATTRIB_KEYINFO_SIZE,
					   &keysize)))
		return 0;

	Tspi_Context_Close(hContext);

	return keysize;
}

char *
tpm_keyring_util_import(char *name)
{
	TSS_HCONTEXT hContext;
	TSS_HKEY hKey;
	TSS_RESULT result;
	TSS_UUID uuid;

	OpenSSL_init();

	/* generate the UUID based on the group name */
	if ((result = generate_uuid(name, &uuid)))
		ReturnError("Error: Failed to generate uuid!");

	if ((result = Tspi_Context_Create(&hContext)))
		ReturnError("Error creating TSP context: %s", Trspi_Error_String(result));

	if ((result = Tspi_Context_Connect(hContext, NULL)))
		ReturnError("Error connecting to TCS: %s", Trspi_Error_String(result));

	/* retrieve the key someone else stored in my system PS */
	if ((result = Tspi_Context_UnregisterKey(hContext, TSS_PS_TYPE_SYSTEM, uuid, &hKey)))
		ReturnError("Error unregistering key from system persistent storage: %s",
			    Trspi_Error_String(result));

	LogDebug("Registered key in user PS");
	if ((result = Tspi_Context_RegisterKey(hContext, hKey, TSS_PS_TYPE_USER, uuid,
					       TSS_PS_TYPE_SYSTEM, SRK_UUID)))
		ReturnError("Error registering key in user persistent storage: %s",
			    Trspi_Error_String(result));

	Tspi_Context_Close(hContext);

	return NULL;
}

char *
tpm_keyring_util_create_group_retry(tpm_keyring_util_ctx *ctx, char *pass)
{
	char *ret;
	TSS_RESULT result;
	TSS_UUID uuid;

	OpenSSL_init();

	if ((result = generate_uuid(ctx->name, &uuid)))
		ReturnError("Failed to generate a new uuid: %s", Trspi_Error_String(result));

	if ((ret = wrap(ctx, pass)))
		return ret;

	LogDebug("Registered key in user PS");
	if ((result = Tspi_Context_RegisterKey(ctx->hContext, ctx->hKey, TSS_PS_TYPE_USER, uuid,
					       TSS_PS_TYPE_SYSTEM, SRK_UUID)))
		ReturnError("Failed to register key: %s", Trspi_Error_String(result));

	Tspi_Context_Close(ctx->hContext);

	return NULL;
}

char *
tpm_keyring_util_create_group(tpm_keyring_util_ctx *ctx)
{
	TSS_RESULT result;
	char *ret;
	TSS_UUID uuid;

	OpenSSL_add_all_algorithms();

	/* generate the UUID based on the group name */
	if ((result = generate_uuid(ctx->name, &uuid)))
		ReturnError("Error: Failed to generate uuid!");

	if ((ret = generate_swkey(ctx)))
		return ret;

	if ((ret = save_swkey(ctx)))
		return ret;

	if ((ret = init_tss(ctx)))
		return ret;

	if ((ret = wrap(ctx, NULL)))
		return ret;

	LogDebug("Registered key in user PS");
	if ((result = Tspi_Context_RegisterKey(ctx->hContext, ctx->hKey, TSS_PS_TYPE_USER, uuid,
					       TSS_PS_TYPE_SYSTEM, SRK_UUID))) {
		Tspi_Context_Close(ctx->hContext);
		ReturnError("Error registering key in user persistent storage: %s",
			    Trspi_Error_String(result));
	}

	Tspi_Context_Close(ctx->hContext);

	return NULL;
}

